
//ECE 4760 Final Project
//Haowen Zheng(hz86), 12/20/2013
//Atmega 1284P version

//PARAMETERS:
//PortA for 4X4 keypad,
//PortB.0 for IR output,
//PortC for LCD display, 
//PortD.6 for IR input

#define F_CPU 16000000UL
#include <avr/io.h>
#include <util/delay.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <avr/eeprom.h>  
#include <avr/pgmspace.h>
#include <stdlib.h>
#include <string.h>

#include "uart.h"
#include "lcd_lib.h"

#define maxkeys 16
#define PORTDIR DDRA
#define PORTDATA PORTA
#define PORTIN PINA

#define NoPush 1 
#define MaybePush 2
#define Pushed 3
#define MaybeNoPush 4

#define Input 0
#define Output 1

#define t_db 50
#define countMS 62
#define BITtime 562


// The raw keyscan
unsigned char key;   
// The decoded button number
unsigned char butnum;


int final_but;           //Button pushed by the user
int prev_but;            //Record previous button        


volatile unsigned int time_db,time_MS, carrier_freq, Counter, bitCount;
unsigned long IRsignal = 0b00000000000000000000000000000000;
unsigned long SentIR;
unsigned long setbit = 0b10000000000000000000000000000000;
unsigned long IRclear = 0b11111111111111111111111111111111;
unsigned char PushState, IRstate;       // state machine 
unsigned char iomode;		  			// input/output mode
unsigned char PushFlag;       			// message indicating a button push 
volatile uint16_t falltime, prev_falltime, fall_diff;

// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

//key pad scan table
unsigned char keytbl[16]={0xee, 0xed, 0xeb, 0xe7, 
                          0xde, 0xdd, 0xdb, 0xd7, 
                          0xbe, 0xbd, 0xbb, 0xb7, 
                          0x7e, 0x7d, 0x7b, 0x77};

// LCD Initialize
const int8_t LCD_initialize[] PROGMEM = " LCD Initialized\0";
const int8_t LCD_input[] PROGMEM = " Wait For Input  \0";
const int8_t LCD_recorded[] PROGMEM = " Signal Recorded!\0";
const int8_t LCD_output[] PROGMEM = " Select Button   \0";
const int8_t LCD_sending[] PROGMEM = " Sending Signal  \0";
const int8_t LCD_detected[] PROGMEM = " Signal Detected  \0";
int8_t lcd_buffer[17];		// LCD display buffer


//*****************************************************************************************

ISR (TIMER0_OVF_vect)
{//Decrement the three times if they are not already zero
     if (time_MS==0){
       if (time_db>0) --time_db;
	   time_MS=countMS;
     }
     time_MS--;
}     

void resetIR(void)
{
	IRstate = 0;
	prev_falltime = 0;
	falltime = 0;
	bitCount = 0;
}

// Falling edge interrupt handler
ISR (TIMER1_CAPT_vect)
{
	 prev_falltime = falltime;
	 falltime = ICR1;
	 fall_diff = falltime - prev_falltime;
	 switch (IRstate)
          {
		  	 case 0:
			 	IRstate = 1;
				falltime = 0;
				prev_falltime = 0;
				TCNT1 = 0x00;
				break;
             case 1: 
			 	// Head pulse, between 13000/4 and 14000/4
                if ((fall_diff>=3250) && (fall_diff<=3500))
				{
                    IRstate = 2;
                }
                else{
                    resetIR();
                }
                break;
				// Recording data
             case 2:
			 	if (bitCount < 32)
				{
			 		// bit is 1, between 2000/4 and 2500/4
                	if ((fall_diff>=500) && (fall_diff<=625)) 
                	{
                	    IRsignal |= setbit;
                	}
					// bit is 0, between 1000/4 and 1500/4
					else if ((fall_diff>=250) && (fall_diff<=375))
					{
						IRsignal &= ~setbit;
					}
					else{
						resetIR();
					}
					// update count and shift setbit
					bitCount++;
					setbit >>= 1;
				}
                else {
                    resetIR();
					TIMSK1 &= ~0x20;
                }
                break;
             
			 default: resetIR();
			 	break;
          }

}

//*****************************************************************************************
//function to scan keypad
int keypad(void){
  //get lower nibble
    PORTDIR = 0x0f;
    PORTDATA = 0xf0; 
    _delay_us(5);
    key = PORTIN;
      
    //get upper nibble
    PORTDIR = 0xf0;
    PORTDATA = 0x0f; 
    _delay_us(5);
    key = key | PORTIN;
      
    //find matching keycode in keytbl
    if (key != 0xff){

      for (butnum=0; butnum<maxkeys; butnum++)
      {  
        if (keytbl[butnum]==key)  break;   
      }

      if (butnum==maxkeys) butnum=0;
      else butnum++;     //adjust by one to make range 1-16
     
    }    
    else butnum=0;                
     
    return butnum;
}              

//*****************************************************************************************
void debounce(void){
     time_db = t_db;            //restore cycle time
     final_but = keypad();
		  //fprintf(stdout,"%d \n\r", final_but);
          switch (PushState)
          {
             case NoPush: 
                if (final_but==prev_but) 
				{
                    PushState=MaybePush;
                }
                else {
                    PushState=NoPush;
                    prev_but=final_but;
                }
                break;
             case MaybePush:
                if (final_but==prev_but) 
                {
                   PushState=Pushed;   
                }
                else {
                    PushState=NoPush;
                    prev_but=final_but;
               }
                break;
             case Pushed:  
                if (final_but==prev_but) 
				{
                    PushState=Pushed;
					PushFlag=1;
					Counter = Counter + 1;
                } 
                else {
                    PushState=MaybeNoPush;  
                    prev_but=final_but;
					PushFlag=0;
					Counter = -1;
               }
                break;
             case MaybeNoPush:
                if (final_but==prev_but) 
				{
                    PushState=Pushed; 
                }
                else {
                   PushState=NoPush;
                   PushFlag=0;
                   prev_but=final_but;
                }    
                break;
          }
}

//*****************************************************************************************
void IRdecode(void)
{
	// Enable and initialize interrupt
	setbit = 0b10000000000000000000000000000000;
	TIMSK1 |= 0x20;
	TCNT1 = 0x00;
	resetIR();
}

//*****************************************************************************************
void IRcarrier(unsigned int IRtime)
{
	// Create 38kHz wave
	int i;
	for (i = 0; i < (IRtime/13); i++){
		PORTB ^= 0x01;
		_delay_us(12);
	}
}

void IRsendSignal(unsigned long code)
{
	int i;

	// send leader code
	IRcarrier(9000);
	_delay_us(4500);

	// send the rest code
	for (i = 0; i < 32; i++){
		IRcarrier(BITtime);
		if (code & 0x80000000) _delay_us(3*BITtime);
		else _delay_us(BITtime);
		code <<= 1;
	}
	IRcarrier(BITtime);
}

//*****************************************************************************************
void initialize(void){
     //Initialize PORTS

     DDRD = 0x02;            // Set up PORTD for UART D.1, set up D.2 for LED
                             // keypad port PORTA initlized in keypad()
	 DDRB = 0x01;
	 PORTB = 0;

     //Initialize timers
     //set up timer 0 for 1 mSec timebase 
     TIMSK0 = (1<<TOIE0);   //turn on timer 0 overflow ISR 
     //set prescalar to divide by 1 
     TCCR0B = 1;   
     // Set OC0A on Compare Match, use fast PWM
     TCCR0A = (1<<COM0A1) | (1<<COM0A0) | (1<<WGM00) | (1<<WGM01); 

	 // set up timer 1, prescaler set to 64
	 //TCCR1B = (1<<ICES1) | (1<<CS11) | (1<<CS10);
	 TCCR1B |= 0x03;
	 TCCR1B &= ~0x40;

     //init the UART -- uart_init() is in uart.c
     uart_init();
     stdout = stdin = stderr = &uart_str;
     fprintf(stdout,"Starting...\n\r");

	 // LCD init
	 LCDinit();	//initialize the display
	 LCDcursorOFF();
	 LCDclr();				//clear the display
	 LCDGotoXY(0,0);
	 CopyStringtoLCD(LCD_initialize, 0, 0);
	 //LCDclr();
	 
     //Initialize variables
     final_but = 0;
     prev_but = -1;

	 falltime = 0;
	 prev_falltime = 0;
	 bitCount = 0;

     time_db = t_db;
     time_MS = countMS;   //setup 1ms time scale

     PushState = NoPush;  
	 IRstate = 0;  
     PushFlag = 0;
	 Counter = -1;
	 iomode = -1;

     sei();
}

//*****************************************************************************************
int main(void){
     initialize();

     while(1){
        if (time_db == 0) debounce();
	 
	 // Prompt user to input signal
	 	if (final_but == 4){
	 		CopyStringtoLCD(LCD_input, 0, 0);
			iomode = Input;
			IRdecode();
			
	 	}
		if (final_but == 8){
	 		CopyStringtoLCD(LCD_output, 0, 0);
			iomode = Output;
	 	}
		if ((iomode == Input) && (((final_but>0) && (final_but<4)) || ((final_but>4) && (final_but<8)) || ((final_but>8) && (final_but<12)))){
			// Store signal to the pushed button
			switch (final_but){
				// Store IRsignal to EEPROM according to the button pushed
				case 1:
					eeprom_write_block((const void *)&IRsignal, (void *)0, 4);
					break;
				case 2:
					eeprom_write_block((const void *)&IRsignal, (void *)4, 4);
					break;
				case 3:
					eeprom_write_block((const void *)&IRsignal, (void *)8, 4);
					break;
				case 5:
					eeprom_write_block((const void *)&IRsignal, (void *)12, 4);
					break;
				case 6:
					eeprom_write_block((const void *)&IRsignal, (void *)16, 4);
					break;
				case 7:
					eeprom_write_block((const void *)&IRsignal, (void *)20, 4);
					break;
				case 9:
					eeprom_write_block((const void *)&IRsignal, (void *)24, 4);
					break;
				case 10:
					eeprom_write_block((const void *)&IRsignal, (void *)28, 4);
					break;
				case 11:
					eeprom_write_block((const void *)&IRsignal, (void *)32, 4);
					break;
			}
			CopyStringtoLCD(LCD_recorded, 0, 0);
		}
		if ((iomode == Output) && (((final_but>0) && (final_but<4)) || ((final_but>4) && (final_but<8)) || ((final_but>8) && (final_but<12)))){
			// Send signal
			CopyStringtoLCD(LCD_sending, 0, 0);
			// Get corresponding signal from EEPROM according to the button pushed
			switch (final_but){
				case 1:
					eeprom_read_block((void *)&SentIR, (const void *)0, 4);
					break;
				case 2:
					eeprom_read_block((void *)&SentIR, (const void *)4, 4);
					break;
				case 3:
					eeprom_read_block((void *)&SentIR, (const void *)8, 4);
					break;
				case 5:
					eeprom_read_block((void *)&SentIR, (const void *)12, 4);
					break;
				case 6:
					eeprom_read_block((void *)&SentIR, (const void *)16, 4);
					break;
				case 7:
					eeprom_read_block((void *)&SentIR, (const void *)20, 4);
					break;
				case 9:
					eeprom_read_block((void *)&SentIR, (const void *)24, 4);
					break;
				case 10:
					eeprom_read_block((void *)&SentIR, (const void *)28, 4);
					break;
				case 11:
					eeprom_read_block((void *)&SentIR, (const void *)32, 4);
					break;
			}
			fprintf(stdout,"%lx \n\r", SentIR);
			if (Counter == 0) IRsendSignal(SentIR);
		}
		if (IRstate == 2){
			// Display Signal detected message
			CopyStringtoLCD(LCD_detected, 0, 0);
		}
		if (final_but == 13){
			// Clear I/O mode
			iomode = -1;
			// Clear all the stored data and go back to initial state
			eeprom_write_block((const void *)&IRclear, (void *)0, 4);
			eeprom_write_block((const void *)&IRclear, (void *)4, 4);
			eeprom_write_block((const void *)&IRclear, (void *)8, 4);
			eeprom_write_block((const void *)&IRclear, (void *)12, 4);
			eeprom_write_block((const void *)&IRclear, (void *)16, 4);
			eeprom_write_block((const void *)&IRclear, (void *)20, 4);
			eeprom_write_block((const void *)&IRclear, (void *)24, 4);
			eeprom_write_block((const void *)&IRclear, (void *)28, 4);
			eeprom_write_block((const void *)&IRclear, (void *)32, 4);
			CopyStringtoLCD(LCD_initialize, 0, 0);
		}
	}
}


